/*
    This file is part of autower, Copyright (C) Christoph Langguth, 2006

    autower is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    autower is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with autower; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/
/* Version history

1.01 - November 28, 2006
* fixed reading of the initialization file for the command-line version. I knew
  something would go wrong :-/

1.02 - November 29, 2006
* Corrected the connection method to FSUIPC for the DLL version to use FSUIPC_Open2

1.03 - December 3, 2006
* File parsing is now done using memory mapped files. This gives a speedup of about 10%.
* Added the graphical progress display when building datafile.
* Added possibility for the simple fixes for airports with a single runway.

*/






// Oh, don't blame me for some strange or unusual things in the code,
// being a Java programmer I'm not that much into C...

#include <windows.h>
#include <commctrl.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "autower.h"
#include "FSUIPC_User.h"

#define DISPLAY_DEBUG 5
#define DISPLAY_DETAIL 4
#define DISPLAY_INFO 3
#define DISPLAY_WARN 2
#define DISPLAY_ERROR 1
#define DISPLAY_FATAL 0

#define ICAOCODE = " 0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
#define RADIUS 3443.9
#define PI 3.141592653589793238462643383279502884197169399375105820974944592307816

/* convenience macros */
#define READ_WORD(VAR) memcpy(&VAR,filePointer,2); filePointer += 2;
#define READ_DWORD(VAR) memcpy(&VAR,filePointer,4); filePointer += 4;
#define READ_BYTE(VAR) memcpy(&VAR,filePointer++,1);
#define DEG2RAD(VAR) (((float)((int)VAR))*PI)/180.0




// This stuff is required for building the DLL only.
// Pass the -DBUILD_DLL option to the compiler.
// Thanks to some guy who posted this on the RosAsm forum - I wouldn't have found this out by myself...
#ifdef BUILD_DLL
typedef int ID;
typedef unsigned int VAR32;
typedef VAR32 FLAGS32;
#define   FSAPI __stdcall

// These are the functions called when loading and unloading the module
//
void FSAPI module_init ();

void FSAPI module_deinit () {}

// These are the two structs that are exported from the DLL
//
typedef struct IMPORTTABLE
{
    struct
    {
        ID fnID;
        PVOID fnPtr;
    } dummy;
} IMPORTTABLE;
__declspec (dllexport) IMPORTTABLE ImportTable;

typedef struct LINKAGE
{
    ID ModuleID;
    void (FSAPI *ModuleInit) ();
    void (FSAPI *ModuleDeInit) ();
    FLAGS32 ModuleFlags;
    UINT32 ModulePriority;
    UINT32 ModuleVersion;
    LPVOID* Table[];
} LINKAGE;
__declspec (dllexport) LINKAGE Linkage;

// The global exported LINKAGE object
//
LINKAGE Linkage =
{
    0x000002d9,
    module_init,
    module_deinit,
    0,
    0,
    0x900, //0x800,
    { 0 }
};

// The global exported IMPORTTABLE object
//
IMPORTTABLE ImportTable =
{
    {
        0x00000000, NULL
    }
}; 
#endif


#ifdef BUILD_DLL
BYTE dllIpcMem[1024];
#define FSUIPC_OPEN FSUIPC_Open2(SIM_ANY, &ipcResult, dllIpcMem, sizeof(dllIpcMem))
#else
#define FSUIPC_OPEN FSUIPC_Open(SIM_ANY, &ipcResult)
#endif






/* user-definable variables */

unsigned int displayLevel = DISPLAY_INFO;

long maxDisplayInfoAltitude = DEFAULT_MAXDISPLAYINFOALTITUDE;
unsigned int towerAlt1Rw = DEFAULT_TOWERALT1RW; 
unsigned int towerAlt2Rw = DEFAULT_TOWERALT2RW;
unsigned int towerAlt3Rw = DEFAULT_TOWERALT3RW;
unsigned int towerAlt4Rw = DEFAULT_TOWERALT4RW;
unsigned int towerAlt5Rw = DEFAULT_TOWERALT5RW;
unsigned int towerAlt6Rw = DEFAULT_TOWERALT6RW;

unsigned int towerMinCandidates = DEFAULT_TOWERMINCANDIDATES;
unsigned int comLockThreshold = DEFAULT_COMLOCKTHRESHOLD;
unsigned int zoomLevel = DEFAULT_ZOOMLEVEL;
unsigned int updateInterval = DEFAULT_UPDATEINTERVAL;
unsigned int simpleFixHeading = DEFAULT_SIMPLEFIXHEADING;
int simpleFixDistance = DEFAULT_SIMPLEFIXDISTANCE;

int displayDuration = DEFAULT_DISPLAYDURATION;

/* global variables */
char buf[MAX_PATH*3] = {0};
char* baseDir;
char* dataFile;
char* iniFile;

LPVOID filePointer;

KdTree kdTree = NULL;

int searchResultCount = 0;
AirportInfo** searchResults = NULL;
double currentLatitude = 0;
double currentLongitude = 0;
double currentAltitude = 0;
WORD currentActiveCom = 0;
AirportInfo* currentAirport = NULL;

/* data file */
HANDLE hData = NULL;

DWORD ipcResult = 0;

/* progress window */
HWND hwndProgressWindow = 0;
HWND hwndProgressBar = 0;
HWND hwndFileCount = 0;
HWND hwndAirportCount = 0;
HWND hwndProgressStatus = 0;

int progressTotalLayers = 0;
int progressLayersDone = 0;

char *ipcErrors[] = {
	"Okay",
	"Attempt to Open when already Open",
	"Cannot link to FSUIPC or WideClient",
	"Failed to Register common message with Windows",
	"Failed to create Atom for mapping filename",
	"Failed to create a file mapping object",
	"Failed to open a view to the file map",
	"Incorrect version of FSUIPC, or not FSUIPC",
	"Sim is not version requested",
	"Call cannot execute, link not Open",
	"Call cannot execute: no requests accumulated",
	"IPC timed out all retries",
	"IPC sendmessage failed all retries",
	"IPC request contains bad data",
	"Maybe running on WideClient, but FS not running on Server, or wrong FSUIPC",
	"Read or Write request cannot be added, memory for Process is full",
};


/* Prototypes */
void display(LPSTR message, int level);
int rebuildDatabase(LPSTR baseDirectory);
int rebuildDatabaseStart();
int checkBglFile(LPSTR bglFile);
int checkBgl0201File(HANDLE hFile, LPSTR bglFile);
int parse0201Airports(HANDLE hFile, LPSTR bglFile, LONG offset, int records);
void dumpAirportInfo(AirportInfo *airport, int displaylevel);
double convertAltitude(DWORD dword);
double convertLatitude(DWORD dword);
double convertLongitude(DWORD dword);

/* Error messages */

LPVOID getErrorMsg() {
	LPVOID lpMsgBuf;
	FormatMessage( 
		FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
		NULL,
		GetLastError(),
		MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
		(LPTSTR) &lpMsgBuf,
		0,
		NULL 
	);
	return lpMsgBuf;
}

/* Progress-window related stuff */

LRESULT CALLBACK progressWindowProcedure(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam) {
	static int cxChar, cyChar;
	switch(message) {
	case WM_CREATE:
		cxChar = LOWORD(GetDialogBaseUnits());
		cyChar = HIWORD(GetDialogBaseUnits());
		InitCommonControls();
		
		hwndProgressStatus = CreateWindow(
			"Static",
			"Please wait for autower to build its database; don't fly yet!",
			WS_CHILD | WS_VISIBLE ,
			0,
			0,
			50 * cxChar,
			cyChar,
			hwnd,
			NULL,
			NULL,
			NULL
		);

		hwndFileCount = CreateWindow(
			"Static",
			"Files scanned: 0",
			WS_CHILD | WS_VISIBLE ,
			0,
			2 * cyChar,
			50 * cxChar,
			cyChar,
			hwnd,
			NULL,
			NULL,
			NULL
		);

		hwndAirportCount = CreateWindow(
			"Static",
			"Airports found: 0",
			WS_CHILD | WS_VISIBLE ,
			0,
			3 * cyChar,
			50 * cxChar,
			cyChar,
			hwnd,
			NULL,
			NULL,
			NULL
		);

		hwndProgressBar = CreateWindowEx(
			0,
			PROGRESS_CLASS,
			"",
			WS_CHILD | WS_VISIBLE | PBS_SMOOTH,
			0,
			4 * cyChar,
			50 * cxChar,
			cyChar,
			hwnd,
			NULL,
			NULL,
			NULL
		);
		return 0;
	case WM_DESTROY:
		PostQuitMessage(0);
		return 0;
	}
	return DefWindowProc(hwnd, message, wParam, lParam);
}


DWORD WINAPI createProgressWindow(LPVOID dummy) {
	int cxChar = LOWORD(GetDialogBaseUnits());
	int cyChar = HIWORD(GetDialogBaseUnits());
	MSG msg;
	WNDCLASS wndclass;
	
	wndclass.style = CS_HREDRAW | CS_VREDRAW;
	wndclass.lpfnWndProc = progressWindowProcedure;
	wndclass.cbClsExtra = 0;
	wndclass.cbWndExtra = 0;
	wndclass.hInstance = 0;
	wndclass.hIcon = LoadIcon(NULL, IDI_APPLICATION);
	wndclass.hCursor = LoadCursor(NULL, IDC_ARROW);
	wndclass.hbrBackground = (HBRUSH) COLOR_BACKGROUND;
	wndclass.lpszClassName = WINDOW_TITLE;
	wndclass.lpszMenuName = NULL;
	
	RegisterClass(&wndclass);
	
	HWND hwndFSMain = FindWindow("FS98MAIN",NULL);
	hwndProgressWindow = CreateWindowEx(
		WS_EX_TOPMOST | WS_EX_NOPARENTNOTIFY | WS_EX_TOOLWINDOW,
		WINDOW_TITLE,
		WINDOW_TITLE,
		WS_OVERLAPPED | WS_SIZEBOX,
		0,
		0,
		cxChar*55,
		cyChar*7,
		hwndFSMain,
		NULL,
		NULL,
		NULL
	);
	
	ShowWindow(hwndProgressWindow, SW_SHOWDEFAULT);
	SetForegroundWindow(hwndProgressWindow);
	UpdateWindow(hwndProgressWindow);
	
	while(GetMessage(&msg, NULL, 0, 0))
	{
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}
	return msg.lParam;
}

void incrementFileCount(int inc) {
	static DWORD lastTickCount = 0;
	static int files = 0;
	
	files += inc;
	
	DWORD newTickCount = GetTickCount();
	if (newTickCount >= lastTickCount + 10) {
		sprintf(buf, "Files scanned: %d (%d/%d scenery layers done)", files, progressLayersDone, progressTotalLayers);
		SendMessage(hwndFileCount, WM_SETTEXT, 0, (LPARAM)buf);
	}
	lastTickCount = newTickCount;
}

void incrementAirportCount(int inc) {
	static DWORD lastTickCount = 0;
	static int airports = 0;
	
	airports += inc;
	
	DWORD newTickCount = GetTickCount();
	if (newTickCount >= lastTickCount + 10) {
		sprintf(buf, "Airports found: %d", airports);
		SendMessage(hwndAirportCount, WM_SETTEXT, 0, (LPARAM)buf);
	}
	lastTickCount = newTickCount;
}

double getDistance(double lat1,double lon1,double lat2,double lon2) {
	/* distance in nautical miles */
	double rlat1, rlat2,rlon1,rlon2;
	rlat1 = lat1*PI / 180.0;
	rlat2 = lat2*PI / 180.0;
	rlon1 = lon1*PI / 180.0;
	rlon2 = lon2*PI / 180.0;
	return acos(sin(rlat1)*sin(rlat2) +cos(rlat1)*cos(rlat2)*cos(rlon2-rlon1)) * RADIUS;

}
void resultCallback(AirportInfo* nextResult) {
	searchResults[searchResultCount++] = nextResult;
	nextResult->currentDistance = getDistance(
		currentLatitude,
		currentLongitude,
		convertLatitude(nextResult->latitude),
		convertLongitude(nextResult->longitude));
	//dumpAirportInfo(nextResult);
}

AirportInfo* findClosest(long factor, int minCount) {
	searchResultCount = 0;
	if (factor < 2) factor = 2;
	double dFactor = factor;
	double minLat, minLon, maxLat, maxLon;
	
	dFactor /= 10;
	minLat = currentLatitude - dFactor / 2;
	minLon = currentLongitude - dFactor;
	
	maxLat = currentLatitude + dFactor / 2;
	maxLon = currentLongitude + dFactor;
	

	ItemType min;
	ItemType max;
	min[0] = minLat;
	min[1] = minLon;
	max[0] = maxLat;
	max[1] = maxLon;
	
	long time = GetTickCount();
	PrintRange(min,max,kdTree);
	time = GetTickCount() - time;
	sprintf(buf,"search for (%f,%f)-(%f,%f) (factor %ld) took %ld ms, yielded %d results\n", minLat,minLon,maxLat,maxLon,factor,time, searchResultCount);
	display(buf, DISPLAY_DEBUG);
	
	/* These strange values for longitude are due to the artificial entries */
	if (searchResultCount < minCount && (minLat > -90 || maxLat < 90 || minLon > -360 || maxLon < 360)) {
		if (factor * 2 > factor) // prevent endless loops caused by overflows
		return findClosest(factor * 2, minCount);
	}
	
	int i;
	AirportInfo *closest = NULL;
	BOOL comLocked = FALSE;
	for(i=0; i < searchResultCount; i++) {
		AirportInfo *checked = searchResults[i];
		BOOL comOk = FALSE;
		// check the COM frequencies of this airport, but only if it's within COM_LOCK_THRESHOLD nm
		if (checked->currentDistance < comLockThreshold) {
			int c;
			for (c = 0; c < AIRPORT_COM_SIZE; c++) {
				WORD freq = checked->COM[c];
				if (freq == 0) break;
				if (freq % 10 == 5) freq -= 5; // hack because the last digit is not used in FS
				if (freq == currentActiveCom) {
					if (currentAirport != checked) {
						sprintf(buf, "COM Frequency matched for %s at distance %f\n", checked->ICAO, checked->currentDistance);
						display(buf,DISPLAY_DEBUG);
					}
					comOk = TRUE;
					break;
				}
			}
		}

		if (closest == NULL || (comOk && !comLocked)) {
			// this airport is better than the previous, either because there was no previous ;-)
			// or because its COM frequency matches while the previous doesn't
			closest = checked;
			comLocked = comOk;
			continue;
		}
		
		if (!comOk && comLocked) {
			// this airport is worse irregardless of its distance
			continue;
		}
		
		// airports are equal wrt to radio availability - let the distance decide.
		if (closest-> currentDistance > checked->currentDistance) {
			comLocked = comOk;
			closest = checked;
		}
	}
	return closest;
}

BOOL registerAirport(AirportInfo *airport, BOOL checkDuplicates, int displaylevel) {
	
	/* insert node into KdTree */
	{
		ItemType location;
		location[0] = convertLatitude(airport->latitude);
		location[1] = convertLongitude(airport->longitude);
		
		/* check if airport already exists. A reasonable assumption will be that the first airport around is ok*/
		currentLatitude = location[0];
		currentLongitude = location[1];
		dumpAirportInfo(airport, displaylevel);

		if (checkDuplicates && kdTree != NULL) {
			AirportInfo *existing = findClosest(2,1);
			if (existing != NULL && !(strcmp(existing->ICAO,airport->ICAO))) {
				sprintf(buf, "Found an existing entry for %s, skipping\n", airport->ICAO);
				display(buf, displaylevel);
				sprintf(buf, "Original airport:\n");
				display(buf, displaylevel);
				dumpAirportInfo(existing, displaylevel);
				sprintf(buf, "New airport (not overwriting):\n");
				display(buf, displaylevel);
				dumpAirportInfo(airport, displaylevel);
				return FALSE;
			}
		}
		
		kdTree = Insert(location, airport, kdTree);
		/* insert artificial nodes for coping with searches close to the date border */
		location[1] += (location[1] > 0 ? -1 : 1) * 360;
		kdTree = Insert(location, airport, kdTree);
		return TRUE;
	}
}

int readDatabase() {
	DWORD nRead = 0;
	unsigned int count = 0;
	
	int i;
	AirportInfo *airports;

	ReadFile(hData,&count,sizeof(DWORD),&nRead,NULL);
	if (count == 0) {
		CloseHandle(hData);
		return 0;
	}
	if (count >> 24 < AIRPORTINFO_VERSION) {
		CloseHandle(hData);
		sprintf(buf,"A datafile from a previous version of autower has been detected.\nBecause of changes in the datafile layout, it needs to be rebuilt.\nIf you want to take advantage of new options, please make sure you're using an up-to-date configuration file.");
		display(buf, DISPLAY_WARN);
		return 0;
	}
	if (count >> 24 > AIRPORTINFO_VERSION) {
		CloseHandle(hData);
		sprintf(buf,"A datafile from a NEWER version of autower has been detected.\nBecause autower will not be able to work with this file, the datafile has to be rebuilt.\n");
		display(buf, DISPLAY_ERROR);
		return 0;
	}
	count &= 0xFFFFFF;
	
	airports = malloc(count * sizeof(AirportInfo));
	ReadFile(hData, airports, count * sizeof(AirportInfo),&nRead,NULL);
	CloseHandle(hData);
	
	for (i = 0; i < count; i++) {
		registerAirport(&(airports[i]), FALSE, DISPLAY_DEBUG);
	}
	return count;
}

BOOL ipcGetData() {
	if (FSUIPC_OPEN || ipcResult < 2) {
		BOOL success = TRUE;
		BYTE readyToFly = 0;
		BYTE pausedInMenu = 0;
		WORD pausedInGame = 0;
		
		long long latitude = 0;
		long long longitude = 0;
		long long altitude = 0;
		
		BOOL goOn = TRUE;

		success &= FSUIPC_Read(0x3364, 1, &readyToFly, &ipcResult);
		success &= FSUIPC_Read(0x3365, 1, &pausedInMenu, &ipcResult);
		success &= FSUIPC_Read(0x0264, 2, &pausedInGame, &ipcResult);
		success &= FSUIPC_Process(&ipcResult);

		if (!success) FSUIPC_Close();
		
		readyToFly = !readyToFly; // strange semantics here
		goOn = success && readyToFly && !pausedInMenu && !pausedInGame;
		sprintf(buf,"success: %d ready: %d pmenu: %d pgame: %d: goOn %d\n",success, readyToFly, pausedInMenu, pausedInGame, goOn);
		display(buf, DISPLAY_DEBUG);
		
		if (!goOn) return FALSE;
		

		long long towerLat = 0;
		long long towerLon = 0;
		long long towerAlt = 0;
		double cTowerLat = 0;
		double cTowerLon = 0;
		double cTowerAlt = 0;
		int zoom = 0;
		byte radio = 0;
		
		success &= FSUIPC_Read(0x0560, 8, &latitude, &ipcResult);
		success &= FSUIPC_Read(0x0568, 8, &longitude, &ipcResult);
		success &= FSUIPC_Read(0x0570, 8, &altitude, &ipcResult);
		success &= FSUIPC_Read(0x0D50, 8, &towerLat, &ipcResult);
		success &= FSUIPC_Read(0x0D60, 8, &towerAlt, &ipcResult);
		success &= FSUIPC_Read(0x0D58, 8, &towerLon, &ipcResult);
		success &= FSUIPC_Read(0x3122, 1, &radio, &ipcResult);
		success &= FSUIPC_Read(0x8330, 2, &zoom, &ipcResult);
		
		success &= FSUIPC_Process(&ipcResult);

		if (!success) FSUIPC_Close();
		
		currentLatitude = ((double)latitude) * 90.0/(10001750.0*65536*65536);
		currentLongitude = ((double)longitude) *360.0/(65536.0*65536*65536*65536);
		currentAltitude = ((double)altitude)/(65536.0*65536);
		
		cTowerLat = ((double)towerLat) * 90.0/(10001750.0*65536*65536);
		cTowerLon = ((double)towerLon) *360.0/(65536.0*65536*65536*65536);
		cTowerAlt = ((double)towerAlt)/(65536.0*65536);
		if (radio & 128) {
			radio = 1;
		} else if (radio & 64) {
			radio = 2;
		} else {
			radio = 0;
		}
		
		WORD freq = 0;
		if (radio == 1) {
			success &= FSUIPC_Read(0x034E, 2, &freq, &ipcResult);
		} else if (radio == 2) {
			success &= FSUIPC_Read(0x3118, 2, &freq, &ipcResult);
		}	
		success &= FSUIPC_Process(&ipcResult);
		freq = (freq & 0xF) + ((freq >> 4) & 0xF) * 10 + ((freq >> 8) & 0xF) * 100 + ((freq >> 12) & 0xF) * 1000;
		freq = (((DWORD)(freq + 10000)) * 10) - 117000;
		currentActiveCom = freq;
		
		goOn = success;
		sprintf(buf,"goOn: %d clat %f clon %f calt %f radio #%d @%.3f tower %f %f %f, zoom %f\n", goOn, currentLatitude, currentLongitude, currentAltitude, radio, (freq+117000)/1000.0, cTowerLat, cTowerLon, cTowerAlt, ((double)zoom/64.0));
		display(buf, DISPLAY_DEBUG);
		
		if (!goOn) return FALSE;
		
		return TRUE;
	} else {
		sprintf(buf,"FSUIPC connection failed: %s\n",ipcErrors[ipcResult]);
		display(buf, DISPLAY_ERROR);
		return FALSE;
	}
}
BOOL ipcDisplay(char *text, int duration) {
	BOOL success = TRUE;
	char safeText[128] = {0};
	strncpy(safeText,text,127);
	
	if (FSUIPC_OPEN || ipcResult < 2) {
		success &= FSUIPC_Write(0x3380, sizeof(safeText), safeText, &ipcResult);
		success &= FSUIPC_Write(0x32FA, sizeof(duration), &duration, &ipcResult);
		success &= FSUIPC_Process(&ipcResult);
		if (!success) FSUIPC_Close();
		return success;
	} else {
		sprintf(buf,"FSUIPC connection failed: %s\n",ipcErrors[ipcResult]);
		display(buf, DISPLAY_ERROR);
		return FALSE;
	}
	return FALSE;
}

void mainLoop() {
	int i=0;
	AirportInfo *newAirport;
	for (i = 0; i < 20;) {
		if (ipcGetData()) {
			newAirport = findClosest(2, towerMinCandidates);
			if (newAirport != currentAirport) {
				double lon, lat, alt;
				long long ipcLon, ipcLat, ipcAlt;
				currentAirport = newAirport;
				
				lat = convertLatitude(newAirport->tLatitude);
				if (lat == 0.0) lat = convertLatitude(newAirport->latitude);

				lon = convertLongitude(newAirport->tLongitude);
				if (lon == 0.0) lon = convertLongitude(newAirport->longitude);
				
				if (newAirport->runways == 1 && newAirport->singleRwyInfo != 0) {
					//simple fix is to be applied
					double rwyLength = ((int)newAirport->singleRwyInfo)/10;
					double rwyHeading = (newAirport->singleRwyInfo - rwyLength*10);
					double twrHeading = rwyHeading + DEG2RAD(simpleFixHeading);
					if (twrHeading >= 2 *PI) twrHeading -= 2*PI;
					double distance = (simpleFixDistance < 0) ? 0 - simpleFixDistance : rwyLength * simpleFixDistance / 100;
					
					double metersLat = distance * cos(twrHeading);
					double metersLon = distance * sin(twrHeading);
					
					double nmLon = metersLon / 1852.01;
					double nmLat = metersLat / 1852.01;
					
					sprintf(buf,"rwyLength: %f, rwyHdg: %f, twrHdg: %f, distance: %f\n", rwyLength, rwyHeading, twrHeading, distance);
					display(buf,DISPLAY_DEBUG);
					sprintf(buf,"metersLat: %f, metersLon: %f => nmLat %f, nmLon %f\n", metersLat, metersLon, nmLat, nmLon);
					display(buf,DISPLAY_DEBUG);
					
					// adjust the offsets according to the latitude
					// 1 degree = 60 nm at the equator, adjusted by cos(lat) when going towards the poles
					nmLon = nmLon * cos(DEG2RAD(lat));
					nmLat = nmLat * cos(DEG2RAD(lat));

					sprintf(buf,"lat %f lon %f, adjusted nmLat %f, nmLon %f\n", lat, lon, nmLat, nmLon);
					display(buf,DISPLAY_DEBUG);
					lat += nmLat / 60.0;
					lon += nmLon / 60.0;
					
					// just in case
					if (lon < -180.0) lon += 360.0;
					else if (lon > 180.0) lon -= 360.0;

					sprintf(buf,"final position lat %f lon %f\n", lat, lon);
					
				}

				alt = convertAltitude(newAirport->tAltitude);
				if (alt == 0.0) {
					switch (newAirport->runways) {
					case 0:
					case 1:
						alt = convertAltitude(newAirport->altitude) + towerAlt1Rw; break;
					case 2:
						alt = convertAltitude(newAirport->altitude) + towerAlt2Rw; break;
					case 3:
						alt = convertAltitude(newAirport->altitude) + towerAlt3Rw; break;
					case 4:
						alt = convertAltitude(newAirport->altitude) + towerAlt4Rw; break;
					case 5:
						alt = convertAltitude(newAirport->altitude) + towerAlt5Rw; break;
					default:
						alt = convertAltitude(newAirport->altitude) + towerAlt6Rw; break;
					}
				}
				
				ipcLat = (long long)(double)(lat * (10001750.0*65536*65536) / 90.0);
				ipcLon = (long long)(double)(lon * (65536.0*65536*65536*65536) / 360.0);
				ipcAlt = (long long)(double)(alt * (65536.0*65536));
				
				sprintf(buf, "Setting tower to lat %f lon %f alt %f\n", lat, lon, alt);
				display(buf, DISPLAY_DETAIL);

				BOOL success = TRUE;
				
				success &= FSUIPC_Write(0x0D60, 8, &ipcAlt, &ipcResult);
				success &= FSUIPC_Write(0x0D50, 8, &ipcLat, &ipcResult);
				success &= FSUIPC_Write(0x0D58, 8, &ipcLon, &ipcResult);
				success &= FSUIPC_Write(0x8330, 2, &zoomLevel, &ipcResult);
				success &= FSUIPC_Process(&ipcResult);
				if (success && (currentAltitude <= (double)maxDisplayInfoAltitude)) {
					sprintf(buf, "Displaying info because current alt %f <= %ld\n", currentAltitude, maxDisplayInfoAltitude);
					display(buf, DISPLAY_DEBUG);
					sprintf(buf, "Tower @ %s (%s)", newAirport->ICAO, newAirport->Name);
					ipcDisplay(buf, displayDuration);
				}
				else {
					sprintf(buf, "Not displaying info: success (%d) != 1, or currentAltitude(%f) > maxDisplayInfoAltitude(%ld)\n", success, currentAltitude, maxDisplayInfoAltitude);
					display(buf, DISPLAY_DEBUG);
				}
				dumpAirportInfo(newAirport, DISPLAY_DETAIL);
			}
		}
		Sleep(updateInterval);
	}
}

BOOL setup() {
	baseDir = malloc(MAX_PATH);
	memset(baseDir, MAX_PATH,0);
	
	iniFile = malloc(MAX_PATH);
	memset(iniFile, MAX_PATH,0);
	
	dataFile = malloc(MAX_PATH);
	memset(dataFile, MAX_PATH,0);
	
	BOOL success = FALSE;
	int tries = 100;
	while (tries-- > 0) {
		success = (FSUIPC_OPEN || ipcResult < 2);
		if (success) break;
		Sleep(1000);
	}
	if (!success) {
		sprintf(buf,"FSUIPC connection failed: %s\n",ipcErrors[ipcResult]);
		display(buf, DISPLAY_FATAL);
		return FALSE;
	}
	success = FSUIPC_Read(0x3E00,MAX_PATH,baseDir,&ipcResult);
	success &= FSUIPC_Process(&ipcResult);
	if (!success || strlen(baseDir) < 1) {
		sprintf(buf,"Problem determining FS9 installation directory!\n");
		display(buf, DISPLAY_FATAL);
		return FALSE;
	}

	if (baseDir[strlen(baseDir)-1] != '\\') {
		strcpy(&(baseDir[strlen(baseDir)]),"\\");
	}
	#ifdef BUILD_DLL
	sprintf(dataFile,"%s%s%s",baseDir,"modules\\",DATAFILE_REL);
	sprintf(iniFile,"%s%s%s",baseDir,"modules\\",INIFILE_REL);
	#else
	sprintf(dataFile,".\\%s",DATAFILE_REL);
	sprintf(iniFile, ".\\%s",INIFILE_REL);
	#endif

	displayLevel = GetPrivateProfileInt(INI_SECTIONKEY,KEY_LOGLEVEL,displayLevel,iniFile);
	displayDuration = GetPrivateProfileInt(INI_SECTIONKEY,KEY_DISPLAYDURATION,displayDuration,iniFile);
	maxDisplayInfoAltitude = GetPrivateProfileInt(INI_SECTIONKEY,KEY_MAXDISPLAYINFOALTITUDE,maxDisplayInfoAltitude,iniFile);
	towerMinCandidates = GetPrivateProfileInt(INI_SECTIONKEY,KEY_TOWERMINCANDIDATES,towerMinCandidates,iniFile);
	comLockThreshold = GetPrivateProfileInt(INI_SECTIONKEY,KEY_COMLOCKTHRESHOLD,comLockThreshold,iniFile);
	zoomLevel = GetPrivateProfileInt(INI_SECTIONKEY,KEY_ZOOMLEVEL,zoomLevel,iniFile);
	updateInterval = GetPrivateProfileInt(INI_SECTIONKEY,KEY_UPDATEINTERVAL,updateInterval,iniFile);
	towerAlt1Rw = GetPrivateProfileInt(INI_SECTIONKEY,KEY_TOWERALT1RW,towerAlt1Rw,iniFile);
	towerAlt2Rw = GetPrivateProfileInt(INI_SECTIONKEY,KEY_TOWERALT2RW,towerAlt2Rw,iniFile);
	towerAlt3Rw = GetPrivateProfileInt(INI_SECTIONKEY,KEY_TOWERALT3RW,towerAlt3Rw,iniFile);
	towerAlt4Rw = GetPrivateProfileInt(INI_SECTIONKEY,KEY_TOWERALT4RW,towerAlt4Rw,iniFile);
	towerAlt5Rw = GetPrivateProfileInt(INI_SECTIONKEY,KEY_TOWERALT5RW,towerAlt5Rw,iniFile);
	towerAlt6Rw = GetPrivateProfileInt(INI_SECTIONKEY,KEY_TOWERALT6RW,towerAlt6Rw,iniFile);
	simpleFixHeading = GetPrivateProfileInt(INI_SECTIONKEY,KEY_SIMPLEFIXHEADING,simpleFixHeading,iniFile);
	simpleFixDistance = GetPrivateProfileInt(INI_SECTIONKEY,KEY_SIMPLEFIXDISTANCE,simpleFixDistance,iniFile);

	sprintf(buf, "Configuration dump:\nlogLevel=%u \ndisplayDuration=%d\nmaxDisplayInfoAltitude=%ld \ntowerMinCandidates=%u \ncomLockThreshold=%u \nzoomLevel=%u \nupdateInterval=%u \ntowerAlt1Rw=%u \ntowerAlt2Rw=%u \ntowerAlt3Rw=%u \ntowerAlt4Rw=%u \ntowerAlt5Rw=%u \ntowerAlt6Rw=%u simpleFixHeading=%u\n simpleFixDistance=%d\n",
		displayLevel,
		displayDuration,
		maxDisplayInfoAltitude,
		towerMinCandidates,
		comLockThreshold,
		zoomLevel,
		updateInterval,
		towerAlt1Rw,
		towerAlt2Rw,
		towerAlt3Rw,
		towerAlt4Rw,
		towerAlt5Rw,
		towerAlt6Rw,
		simpleFixHeading,
		simpleFixDistance
	);
	display(buf, DISPLAY_DETAIL);
	return TRUE;
}


DWORD WINAPI doMain( LPVOID dummy) {
	if (!setup()) {
		sprintf(buf,"A configuration error occurred. Cannot continue.\n");
		display(buf,DISPLAY_FATAL);
		return 1;
	}
	hData = CreateFile(dataFile, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);
	
	int success = 0;

	if (hData != INVALID_HANDLE_VALUE) {
		success = readDatabase();
		sprintf(buf, "Successfully read %i tower locations from database.\n", success);
		if (success) display(buf, DISPLAY_DETAIL);
	}
	if (success == 0) {
		success = rebuildDatabaseStart();
		sprintf(buf, "Successfully imported %i tower locations.\n", success);
		display(buf, DISPLAY_DETAIL);
	}
	if (success == 0) {
		sprintf(buf,"No airports in database - exiting!\n");
		display(buf, DISPLAY_FATAL);
	}

	searchResults = malloc(sizeof(DWORD)*success * 2); //factor 2 because of artificial entries

	mainLoop();
	FSUIPC_Close();
	return 0;
}

#ifdef BUILD_DLL
void FSAPI module_init () {
	DWORD thread;
	CreateThread(NULL,0,&doMain,NULL,0,&thread);
}

BOOL WINAPI DllMain (HANDLE hDll, DWORD dwReason, LPVOID lpReserved) {
	return TRUE;
}
#else
int WINAPI WinMain (HINSTANCE hInstance, HINSTANCE hPrevInstance, PSTR szCmdLine, int iCmdShow) {
	return doMain(NULL);
}
#endif

int rebuildDatabaseStart() {
	hData = INVALID_HANDLE_VALUE;
	hData = CreateFile(dataFile, GENERIC_WRITE, FILE_SHARE_READ, NULL, CREATE_ALWAYS, 0, NULL);
	unsigned int airports = 0;
	DWORD nRead = 0;
	
	if (hData == INVALID_HANDLE_VALUE) {
		LPTSTR err = getErrorMsg();
		sprintf(buf, "Invalid file handle while opening database file, error was: %s.\n", err);
		display(buf, DISPLAY_FATAL);
		LocalFree(err);
		return 0;
	}
	sprintf(buf, "Constructing airport database. This may take a couple of minutes.\nPlease do NOT start a flight yet, but wait for the next message!\n");
	display(buf, DISPLAY_DETAIL);
	WriteFile(hData,&airports,sizeof(airports),&nRead,NULL);
	
	
	char sceneryCfg[MAX_PATH] = {0};
	strcpy(sceneryCfg,baseDir);
	strcat(sceneryCfg,"scenery.cfg");
	sprintf(buf, "Scenery: %s\n", sceneryCfg);
	display(buf,DISPLAY_DEBUG);
	
	int layers[1000] = {0};
	char areaName[9];
	int area;
	char buf2[MAX_PATH] = {0};
	DWORD dummy;
	
	// create the progress window
	CreateThread(NULL,0,&createProgressWindow,NULL,0,&dummy);

	for (area = 1; area < 1000; area++) {
		sprintf(areaName,"Area.%03d",area);
		int layer = GetPrivateProfileInt(areaName,"Layer",0,sceneryCfg);
		GetPrivateProfileString(areaName,"active","false",buf2,sizeof(buf),sceneryCfg);
		CharLower(buf2);
		layers[area] = (!strcmp("true",buf2)) ? layer : -1;
		if (layers[area] != -1) ++progressTotalLayers;
		sprintf(buf,"area %s -> layer %d, active: %d\n",areaName, layer, (layers[area] != -1));
		display(buf,DISPLAY_DEBUG);
	}
	for (dummy=0; dummy < 100; dummy++) {
		if (hwndProgressBar != 0) break;
		Sleep(50);
	}
	SendMessage(hwndProgressBar, PBM_SETRANGE, 0, MAKELPARAM (0, progressTotalLayers)); 
	SendMessage(hwndProgressBar, PBM_SETSTEP, (WPARAM) 1, 0);
	
	/* assign temporary space for the searches during insertion. 10000 should be far enough. */
	searchResults = malloc(sizeof(DWORD)*10000); 
	/* traverse the layers in *backward* order. This avoids having to overwrite stuff and cluttering */
	/* Note: I know this is not efficient. But hey, it's only a maximum of 1000 values anyway! */
	while (TRUE) {
		int maxLayer=-1;
		int maxValue=-1;
		int i;
		for (i=1; i < 1000; i++) {
			if (layers[i] > maxValue) {
				maxLayer = i;
				maxValue = layers[i];
			}
		}
		if (maxLayer == -1) break;
		layers[maxLayer] = -1;
		
		sprintf(areaName,"Area.%03d",maxLayer);
		GetPrivateProfileString(areaName,"Local","",buf2,sizeof(buf2),sceneryCfg);
		
		if (strlen(buf2) == 0) continue;
		// very simple check, but ought to be sufficient.
		if (strlen(buf2) > 1 && buf2[1] == ':') {
			// buf2 is already an absolute path
		}
		else {
			strcpy(buf, baseDir);
			strcat(buf, buf2);
			if (strlen(buf) >= MAX_PATH - 1) continue;
			strcpy(buf2, buf);
			
		}
		airports += rebuildDatabase(buf2);
		progressLayersDone++;
		SendMessage(hwndProgressBar, PBM_STEPIT, 0, 0); 
	}
	// make the numbers are correct.
	Sleep(100);
	incrementFileCount(0);
	incrementAirportCount(0);
	
	// close the progress window after 5 seconds.
	SendMessage(hwndProgressStatus, WM_SETTEXT, 0, (LPARAM)"Airport Database created.");
	SetForegroundWindow(hwndProgressWindow);
	Sleep(5000);
	SendMessage(hwndProgressWindow, WM_DESTROY, 0, 0); 
	
	
	// no need for this anymore.
	free(searchResults);
	
	// encode version number in airports
	airports |= (AIRPORTINFO_VERSION << 24);
	SetFilePointer(hData,0,NULL,FILE_BEGIN);
	WriteFile(hData,&airports,sizeof(airports),&nRead,NULL);
	CloseHandle(hData);
	airports &= 0xFFFFFF;
	return airports;
}

int rebuildDatabase(LPSTR baseDirectory) {

	char dirFilter[MAX_PATH] = {0};
	char fullFoundFile[MAX_PATH] = {0};
	
	HANDLE hFind = INVALID_HANDLE_VALUE;
	WIN32_FIND_DATA findFileData = {0};
	BOOL isFirstFile = TRUE;
	
	int iSuccess = 0;
	
	sprintf(buf,"Base directory: %s\n",baseDirectory);
	display(buf, DISPLAY_DEBUG);
	
	strncpy(dirFilter, baseDirectory, sizeof(dirFilter) - 3);
	strcat(dirFilter,"\\*");
	
	sprintf(buf,"Filter expression is: %s\n", dirFilter);
	display(buf, DISPLAY_DEBUG);
	
	hFind = FindFirstFile(dirFilter, &findFileData);
	if (hFind == INVALID_HANDLE_VALUE) {
		LPTSTR err = getErrorMsg();
		sprintf(buf, "Invalid file handle while searching for \"%s\", error was: %s.\n", dirFilter, err);
		display(buf, DISPLAY_ERROR);
		LocalFree(err);
		return 0;
	}
	
	while (TRUE) {
		if (isFirstFile) {
			isFirstFile = FALSE;
		} else {
			if (!FindNextFile(hFind, &findFileData)) break;
		}
		if (!strcmp(".", findFileData.cFileName)) continue;
		if (!strcmp("..", findFileData.cFileName)) continue;
		
		if (strlen(baseDirectory) + 1 + strlen(findFileData.cFileName) >= sizeof(fullFoundFile)) {
			sprintf(buf, "Path to \"%s\" is too long, skipping!\n", findFileData.cFileName);
			display(buf, DISPLAY_ERROR);
			continue;
		}
		strcpy(fullFoundFile, baseDirectory);
		strcat(fullFoundFile, "\\");
		strcat(fullFoundFile, findFileData.cFileName);
		
		
		if (FILE_ATTRIBUTE_DIRECTORY & findFileData.dwFileAttributes) {
			/* directory; recurse */
			sprintf(buf, "recursing into directory \"%s\"\n", findFileData.cFileName);
			display(buf, DISPLAY_DEBUG);
			
			int cSuccess = rebuildDatabase(fullFoundFile);
			if (cSuccess == -1) {
				FindClose(hFind);
				return -1;
			}
			iSuccess += cSuccess;
		}
		else {
			/* file; check whether it's a .bgl file */
			incrementFileCount(1);

			CharLower(findFileData.cFileName);
			if (!strcmp(&findFileData.cFileName[strlen(findFileData.cFileName)-3],"bgl")) {
				sprintf(buf, "Found BGL file: %s\n", findFileData.cFileName);
				display(buf, DISPLAY_DEBUG);
				iSuccess += checkBglFile(fullFoundFile);
			}
		}
	}
	return iSuccess;
}


int checkBglFile(LPSTR bglFile) {
	HANDLE hFile = INVALID_HANDLE_VALUE;
	HANDLE hMap = INVALID_HANDLE_VALUE;
	LPVOID fileMap = NULL;
	WORD word;
	int success = 0;
	
	hFile = CreateFile(bglFile, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);
	if (hFile == INVALID_HANDLE_VALUE) {
		LPTSTR err = getErrorMsg();
		sprintf(buf, "Unable to open file \"%s\", error was: %s\n", bglFile, err);
		display(buf, DISPLAY_ERROR);
		LocalFree(err);
		return 0;
	}
	hMap = CreateFileMapping(hFile, NULL, PAGE_READONLY,0,0,NULL);
	if (hMap == INVALID_HANDLE_VALUE) {
		LPTSTR err = getErrorMsg();
		sprintf(buf, "Unable to create file mapping \"%s\", error was: %s\n", bglFile, err);
		display(buf, DISPLAY_ERROR);
		LocalFree(err);
		goto ret;
	}
	fileMap = MapViewOfFile(hMap,FILE_MAP_READ,0,0,0);
	if (fileMap == NULL) {
		LPTSTR err = getErrorMsg();
		sprintf(buf, "Unable to map file \"%s\", error was: %s\n", bglFile, err);
		display(buf, DISPLAY_ERROR);
		LocalFree(err);
		goto ret;
	}
	
	filePointer = fileMap;
	/* check BGL file version */
	
	READ_WORD(word);
	
	if (word == 0x201) {
		success = checkBgl0201File(fileMap,bglFile);
		goto ret;
	}
	
	sprintf(buf, "Unsupported BGL ID: %x\n", word);
	display(buf, DISPLAY_DEBUG);

ret:
	if (fileMap != NULL) UnmapViewOfFile(fileMap);
	if (hMap != INVALID_HANDLE_VALUE) CloseHandle(hMap);
	if (hFile != INVALID_HANDLE_VALUE) CloseHandle(hFile);
	return success;
}


int checkBgl0201File(LPVOID fileMap, LPSTR bglFile) {
	DWORD dword = 0;
	WORD nSections = 0;
	DWORD nSubsections = 0;
	DWORD nRecords = 0;
	
	int success = 0;
	int cSection = 0;
	int cSubsection = 0;
	DWORD sectionPointerOffset = 0;
	DWORD sectionHeaderOffset = 0;
	DWORD cSubsectionFilePointer = 0;
	

	sprintf(buf, "Working on 0201 BGL...\n");
	display(buf, DISPLAY_DEBUG);

	/* Read number of sections */
	filePointer = fileMap + 20;
	READ_WORD(nSections);
	
	sprintf(buf, "Found %d sections in file\n", nSections);
	display(buf, DISPLAY_DEBUG);
	
	for (cSection=0; cSection < nSections; cSection++) {
		/* get type of section */
		sectionPointerOffset = 0x38 + 20 * cSection;
		filePointer = fileMap + sectionPointerOffset;
		READ_DWORD(dword);

		sprintf(buf, "Section %d type is %lx\n", cSection, dword);
		display(buf, DISPLAY_DEBUG);
		
		if (dword != 3) continue;
		
		/* found an airport entry */



		/* Skip unknown DWORD */
		READ_DWORD(dword);
		
		READ_DWORD(nSubsections);
		
		READ_DWORD(sectionHeaderOffset);

		/* read subsections */
		cSubsectionFilePointer = sectionHeaderOffset;
		filePointer = fileMap + cSubsectionFilePointer;
		
		sprintf(buf, "Found %ld subsections in airport section, section header at 0x%lx.\n", nSubsections, sectionHeaderOffset);
		display(buf, DISPLAY_DEBUG);
		
		for (cSubsection=0; cSubsection < nSubsections; cSubsection++) {
			/* skip ID */
			READ_DWORD(dword);
			
			READ_DWORD(nRecords);
			sprintf(buf, "Found %ld records in subsection %d.\n", nRecords, cSubsection);
			display(buf, DISPLAY_DEBUG);
			
			READ_DWORD(dword);
			success += parse0201Airports(fileMap,bglFile,dword,nRecords);
			
			cSubsectionFilePointer += 0x10;
			filePointer = fileMap + cSubsectionFilePointer;
		}
	}
	return success;
}	


void convertIcao(DWORD code, char* icao) {
	int pos = 0;
	
	int ci = 0;
	
	code /= 0x20;
	for (pos=4; pos >= 0; pos--) {
		ci = code % 38;
		code /= 38;
		if (ci > 11) {
			ci = ci - 12 + 0x41;
		}
		else if (ci >= 2) {
			ci = ci -2 + 0x30;
		}
		else { //if (ci == 1) {
			ci = 0x20;
		}
		icao[pos] = ci;
		if (ci == 0) break;
	}
	
	/* shift left the unused spaces */
	ci = 0;
	for (pos=0; pos < 5; pos++) {
		if (icao[pos] == 0x20) ci++;
		else break;
	}
	for (pos=0; pos < 6 - ci; pos++) {
		icao[pos] = icao[ci+pos];
	}
}

int parse0201Airports(LPVOID fileMap, LPSTR bglFile, LONG offset, int records) {

	int success = 0;
	int cRecord = 0;
	
	DWORD nRead = 0;
	DWORD dword = 0;
	WORD word = 0;
	char byte = 0;
	
	LONG recordOffset = 0;
	DWORD lastRecordLength = 0;
	
	recordOffset = offset;
	
	for(cRecord = 0; cRecord < records; cRecord++) {
		AirportInfo *airport = malloc(sizeof(AirportInfo));
		DWORD subRecordOffset = 0x34; // initial pointer to additional records 
		int comCount = 0;
		
		memset(airport,0,sizeof(AirportInfo));
	
		recordOffset = recordOffset + lastRecordLength;
		filePointer = fileMap + recordOffset;
		
		READ_WORD(word);
		if (word != 3) {
			sprintf(buf,"Record %d has ID 0x%x, expected 0x3. Aborting parse of subsection.\n", cRecord, word);
			display(buf, DISPLAY_DEBUG);
			goto ret;
		}
		
		READ_DWORD(lastRecordLength);
		
		filePointer = fileMap + recordOffset + 40;

		READ_DWORD(dword);
		convertIcao(dword, airport->ICAO);

		filePointer = fileMap + recordOffset + 6;
		READ_BYTE(byte);
		airport->runways = byte;
		
		filePointer = fileMap + recordOffset + 12;
		READ_DWORD(dword); airport->longitude = dword;
		READ_DWORD(dword); airport->latitude = dword;
		READ_DWORD(dword); airport->altitude = dword;
		
		READ_DWORD(dword); airport->tLongitude = dword;
		READ_DWORD(dword); airport->tLatitude = dword;
		READ_DWORD(dword); airport->tAltitude = dword;
		
		// parse the subrecords
		while (subRecordOffset < lastRecordLength) {
			WORD subRecordType;
			DWORD subRecordLength;
			
			filePointer = fileMap + recordOffset + subRecordOffset;
			
			READ_WORD(subRecordType);
			
			READ_DWORD(subRecordLength);
			
			subRecordOffset += subRecordLength;
			
			if (subRecordType == 0x19) {
				/* NAME subrecord */
				subRecordLength -= 6;
				if (subRecordLength >= sizeof(airport->Name)-1) subRecordLength = sizeof(airport->Name)-1;
				memcpy(airport->Name,filePointer,subRecordLength);
//				ReadFile(hFile, airport->Name,subRecordLength, &nRead, NULL);
				
				sprintf(buf, "Airport name found: %s\n", airport->Name);
				display(buf, DISPLAY_DEBUG);
			}
			
			else if (subRecordType == 0x04 && airport->runways == 1) {
				/* Runway subrecord */
				DWORD rwyLat,rwyLon;
				float rwyLength, rwyHeading;
				filePointer = fileMap+recordOffset+subRecordOffset - subRecordLength + 20;
				READ_DWORD(rwyLon);
				READ_DWORD(rwyLat);
				
				float apLon = convertLongitude(airport->tLongitude);
				float apLat = convertLatitude(airport->tLatitude);
				if (apLon == 0) apLon = convertLongitude(airport->longitude);
				if (apLat == 0) apLat = convertLatitude(airport->latitude);
				float lonDiff = convertLongitude(rwyLon) - apLon;
				float latDiff = convertLatitude(rwyLat) -apLat;
				if (lonDiff < 0) lonDiff = -lonDiff;
				if (latDiff < 0) latDiff = -latDiff;
				
				if (lonDiff < 1.0/5000 && latDiff < 1.0/5000) {
					sprintf(buf,"lat/lon differences for tower within threshold for single runway center at %s\n", airport->ICAO);
					display(buf, DISPLAY_DEBUG);
					// skip
					READ_DWORD(dword);
					READ_DWORD(rwyLength);
					//skip
					READ_DWORD(dword);
					READ_DWORD(rwyHeading);
					
					float rwyInfo = (((int)rwyLength)*10) + DEG2RAD(rwyHeading);
					airport->singleRwyInfo = rwyInfo;
					sprintf(buf, "Single Rwy Info: (%f,%f), len %f m, hdg %f -> combined %f\n",
						convertLatitude(rwyLat), convertLongitude(rwyLon),
						rwyLength, rwyHeading, rwyInfo);
					display(buf,DISPLAY_DEBUG);
				}
			}
			
			else if (subRecordType == 0x12) {
				/* COM subrecord */
				READ_WORD(word);
				if (word == 0x06) {
					/* Tower COM */
					READ_DWORD(dword);
					sprintf(buf, "Found tower COM #%d for %s at %.3f\n", comCount++, airport->ICAO, (double)dword/1000000.0);
					display(buf, DISPLAY_DEBUG);
					if (comCount >= AIRPORT_COM_SIZE) {
						sprintf(buf, "WARNING: COM #%d *NOT INSERTED* for %s (not enough space)\n", comCount, airport->ICAO);
						display(buf, DISPLAY_WARN);
					}
					else {
						(airport->COM)[comCount-1] = (dword/1000 - 117000);
					}
				}
				else {
					sprintf(buf, "Skipping COM subrecord of COM type 0x%x\n", word);
					display(buf, DISPLAY_DEBUG);
				}
			}
			else {
				sprintf(buf, "Skipping subrecord of type 0x%x\n", subRecordType);
				display(buf, DISPLAY_DEBUG);
			}
		}
		
		if (registerAirport(airport,TRUE, DISPLAY_DETAIL)) {
			if (hData != INVALID_HANDLE_VALUE) {
				WriteFile(hData,airport,sizeof(AirportInfo),&nRead,NULL);
			}
			success++;
			incrementAirportCount(1);
		}
		else {
			/* the airport was not inserted, we can safely free the memory. */
			free(airport);
		}
	}
	
ret:
	return success;
}


double convertAltitude(DWORD dword) {
	return ((double)dword) / ((double)1000);
}

double convertLongitude(DWORD dword) {
	return ((double) dword) * (360.0 / (3* 0x10000000)) - 180.0;
}

double convertLatitude(DWORD dword) {
	//printf("CONVERT: %lx -> %f", dword, 90.0 - ((double)dword) * (180.0 / (2* 0x10000000)));
	return 90.0 - ((double)dword) * (180.0 / (2* 0x10000000));
}

void dumpAirportInfo(AirportInfo *airport, int displaylevel) {
	if (airport == NULL) {
		sprintf(buf, "Nothing to display.\n");
		display(buf, displaylevel);
		return;
	}
	double alt = convertAltitude(airport->altitude);
	double lon = convertLongitude(airport->longitude);
	double lat = convertLatitude(airport->latitude);

	double talt = convertAltitude(airport->tAltitude);
	double tlon = convertLongitude(airport->tLongitude);
	double tlat = convertLatitude(airport->tLatitude);

	sprintf(buf, "%s\t%s: AP @(%f, %f, %f), %d rwys; TWR @(%f, %f, %f) COM @ [%.3f %.3f %.3f %.3f %.3f %.3f %.3f %.3f] dst %f nm\n",
		airport->ICAO,
		airport->Name,
		lat, lon, alt, airport->runways,
		tlat,tlon,talt,
		((float) (airport->COM[0]) + 117000) /1000,
		((float) (airport->COM[1]) + 117000) /1000,
		((float) (airport->COM[2]) + 117000) /1000,
		((float) (airport->COM[3]) + 117000) /1000,
		((float) (airport->COM[4]) + 117000) /1000,
		((float) (airport->COM[5]) + 117000) /1000,
		((float) (airport->COM[6]) + 117000) /1000,
		((float) (airport->COM[7]) + 117000) /1000,
		airport->currentDistance);
	display(buf, displaylevel);
}








void display(LPSTR message, int level) {
	if (level <= displayLevel) {
		#ifdef BUILD_DLL
		if (level <= DISPLAY_ERROR)
			MessageBox(NULL,message,WINDOW_TITLE,MB_ICONERROR);
		else if (level == DISPLAY_WARN) 
			MessageBox(NULL,message,WINDOW_TITLE,MB_ICONWARNING);
		else
			MessageBox(NULL,message,WINDOW_TITLE,MB_ICONINFORMATION);
		#else
		printf(message);
		#endif
	}
}


